# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
from pkg_resources import get_distribution

sys.path.insert(0, os.path.abspath('..'))

import loam

html_theme = 'sphinx_rtd_theme'

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '4.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
]

root_doc = 'index'

autodoc_member_order = 'bysource'
autoclass_content = 'both'

# -- Project information -----------------------------------------------------
project = 'loam'
copyright = '2018 - 2022, Adrien Morison'
author = 'Adrien Morison'

# The full version, including alpha/beta/rc tags.
release = get_distribution("loam").version
# The short X.Y version.
version = '.'.join(release.split('.')[:2])

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Options for HTML output ----------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'loamdoc'


# -- Options for LaTeX output ---------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (root_doc, 'loam.tex', 'loam Documentation',
     'Adrien Morison', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (root_doc, 'loam', 'loam Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (root_doc, 'loam', 'loam Documentation',
     author, 'loam', 'One line description of project.',
     'Miscellaneous'),
]



from dataclasses import dataclass
from pathlib import Path

import pytest

from loam.cli import Subcmd, CLIManager
from loam.tools import switch_opt, path_entry
from loam.base import entry, Section, ConfigBase


@dataclass
class SecA(Section):
    optA: int = entry(val=1, doc="AA", cli_short='a')
    optB: int = entry(val=2, doc="AB", in_file=False)
    optC: int = entry(val=3, doc="AC")
    optBool: bool = switch_opt(True, 'o', 'Abool')


@dataclass
class SecB(Section):
    optA: int = entry(val=4, doc="BA")
    optB: int = entry(val=5, doc="BB", in_file=False)
    optC: int = entry(val=6, doc="BC", in_cli=False)
    optBool: int = switch_opt(False, 'o', 'Bbool')


@dataclass
class Conf(ConfigBase):
    sectionA: SecA
    sectionB: SecB


@pytest.fixture
def conf() -> Conf:
    return Conf.default_()


@pytest.fixture(params=['subsA'])
def sub_cmds(request):
    subs = {}
    subs['subsA'] = {
        'common_': Subcmd('subsA loam test'),
        'bare_': Subcmd(None, 'sectionA'),
        'sectionB': Subcmd('sectionB subcmd help'),
    }
    return subs[request.param]


@pytest.fixture
def climan(conf, sub_cmds):
    return CLIManager(conf, **sub_cmds)


@pytest.fixture
def cfile(tmp_path):
    return tmp_path / 'config.toml'


@dataclass
class SectionA(Section):
    some_n: int = 42
    some_str: str = "foo"


@pytest.fixture
def section_a() -> SectionA:
    return SectionA()


@dataclass
class SectionB(Section):
    some_path: Path = path_entry(".", "")
    some_str: str = entry(val="bar", in_file=False)


@pytest.fixture
def section_b() -> SectionB:
    return SectionB()


@dataclass
class SectionNotInFile(Section):
    some_int: int = entry(val=0, in_file=False)
    some_str: str = entry(val="baz", in_file=False)


@dataclass
class MyConfig(ConfigBase):
    section_a: SectionA
    section_b: SectionB
    section_not_in_file: SectionNotInFile


@pytest.fixture
def my_config() -> MyConfig:
    return MyConfig.default_()



"""Define entries for common containers."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, TypeVar, Callable, Optional, Tuple

from .base import Entry


T = TypeVar("T")
U = TypeVar("U")


@dataclass(frozen=True)
class TupleEntry(Generic[T]):
    """Represent a tuple[T, ...] entry.

    It is able to:

    - parse a TOML array [elt_a, elt_b, elt_c] into a tuple[T, ...] by feeding
      each element to :attr:`inner_from_toml`.
    - parse a string into a tuple[T, ...] by splitting the string with
      :attr:`str_sep`, stripping each substrings, and feeding those to
      :attr:`inner_from_toml` (which therefore should be able to parse strings
      into T). Set :attr:`str_sep` to "" to split on whitespaces, and set it to
      None to raise a TypeError when attempting to parse a string into a tuple.
    - dump the tuple as a TOML array using `inner_to_toml` to convert each
      element into a type with a TOML representation.

    Note that you don't have to specify `inner_to_toml` if the inner type is
    already representable as a toml object.
    """

    inner_from_toml: Callable[[Any], T]
    inner_to_toml: Optional[Callable[[T], object]] = None
    str_sep: Optional[str] = ","

    @staticmethod
    def wrapping(
        tuple_entry: TupleEntry[U], str_sep: Optional[str]
    ) -> TupleEntry[Tuple[U, ...]]:
        """Produce a tuple entry wrapping another one.

        It is the user responsibility to check that separators are different in
        nested TupleEntries.
        """
        return TupleEntry(
            inner_from_toml=tuple_entry.from_toml,
            inner_to_toml=tuple_entry.to_toml,
            str_sep=str_sep,
        )

    def entry(self, default: object = (), doc: str = "", in_file: bool = True,
              in_cli: bool = True, cli_short: Optional[str] = None,
              cli_zsh_comprule: Optional[str] = "") -> Tuple[T, ...]:
        """Produce a :class:`dataclasses.Field` with desired options.

        See :class:`~loam.base.Entry` for an explanation on the parameters.
        """
        return Entry(
            val=self.from_toml(default),
            doc=doc,
            from_toml=self.from_toml,
            to_toml=self.to_toml,
            in_file=in_file,
            in_cli=in_cli,
            cli_short=cli_short,
            cli_kwargs={"nargs": "?", "const": ""} if in_cli else {},
            cli_zsh_comprule=cli_zsh_comprule,
        ).field()

    def from_toml(self, obj: object) -> Tuple[T, ...]:
        """Build a tuple from a TOML object."""
        if isinstance(obj, str):
            sep = self.str_sep
            if sep is None:
                raise TypeError(
                    "Cannot parse str into a Tuple as str_sep is None")
            sep = sep if sep != "" else None
            obj = tuple(map(str.strip, obj.split(sep)))
        if isinstance(obj, (list, tuple)):
            return tuple(map(self.inner_from_toml, obj))
        raise TypeError(
            f"obj should be a str, tuple, or list; got a {obj.__class__}")

    def to_toml(self, val: Tuple[T, ...]) -> Tuple[object, ...]:
        """Transform into a TOML array."""
        if self.inner_to_toml is None:
            return val
        return tuple(map(self.inner_to_toml, val))


@dataclass
class MaybeEntry(Generic[T]):
    """Represent an Optional[T] entry."""

    inner_from_toml: Callable[[Any], T]
    inner_to_toml: Optional[Callable[[T], object]] = None
    none_to_toml: object = ""

    def entry(self, default: object = None, doc: str = "",
              in_file: bool = True, in_cli: bool = True,
              cli_short: Optional[str] = None,
              cli_zsh_comprule: Optional[str] = "") -> Optional[T]:
        """Produce a :class:`dataclasses.Field` with desired options.

        See :class:`~loam.base.Entry` for an explanation on the parameters.
        """
        return Entry(
            val_factory=lambda: self.from_toml(default),
            doc=doc,
            from_toml=self.from_toml,
            to_toml=self.to_toml,
            in_file=in_file,
            in_cli=in_cli,
            cli_short=cli_short,
            cli_kwargs={"nargs": "?", "const": None} if in_cli else {},
            cli_zsh_comprule=cli_zsh_comprule,
        ).field()

    def from_toml(self, obj: object) -> Optional[T]:
        """Build an Optional[T] from a TOML object."""
        if obj is None or self.none_to_toml == obj:
            return None
        return self.inner_from_toml(obj)

    def to_toml(self, val: Optional[T]) -> object:
        """Transform into a TOML object."""
        if val is None:
            return self.none_to_toml
        if self.inner_to_toml is not None:
            return self.inner_to_toml(val)
        return val



from dataclasses import dataclass
from pathlib import Path
from shlex import split as shsplit
from typing import Optional, Tuple

import pytest

from loam.base import Section, ConfigBase
from loam.cli import CLIManager, Subcmd
from loam.collections import MaybeEntry, TupleEntry


@pytest.fixture
def tpl():
    return TupleEntry(int)


@pytest.fixture
def maybe_int():
    return MaybeEntry(int, none_to_toml="none")


@pytest.fixture
def maybe_path():
    return MaybeEntry(Path, str)


@dataclass
class Sec(Section):
    tpl: Tuple[int] = TupleEntry(int).entry()
    mfloat: Optional[float] = MaybeEntry(float).entry()
    mpath: Optional[Path] = MaybeEntry(Path, str).entry()


@dataclass
class Config(ConfigBase):
    sec: Sec


@pytest.fixture
def conf() -> Config:
    return Config.default_()


@pytest.fixture
def climan(conf) -> CLIManager:
    return CLIManager(conf, bare_=Subcmd("", "sec"))


def test_tuple_entry_int(tpl):
    assert tpl.from_toml("5, 6,7 ,1") == (5, 6, 7, 1)
    assert tpl.to_toml((3, 4, 5)) == (3, 4, 5)


def test_tuple_entry_from_invalid_type(tpl):
    with pytest.raises(TypeError):
        tpl.from_toml(8)


def test_tuple_entry_whitespace():
    tpl = TupleEntry(inner_from_toml=int, str_sep="")
    assert tpl.from_toml("5  6 7\t1\n  42") == (5, 6, 7, 1, 42)


def test_tuple_entry_from_arr_str(tpl):
    assert tpl.from_toml(["5", "3", "42"]) == (5, 3, 42)


def test_tuple_entry_from_str_no_sep():
    tpl = TupleEntry(inner_from_toml=int, str_sep=None)
    with pytest.raises(TypeError):
        tpl.from_toml("42,41")
    tpl.from_toml([42, 41]) == (42, 41)


def test_tuple_entry_path():
    root = Path("path")
    tpl = TupleEntry(inner_from_toml=Path, inner_to_toml=str)
    assert tpl.from_toml(["path/1", "path/2"]) == (root / "1", root / "2")
    assert tpl.to_toml((root, root / "subdir")) == ("path", "path/subdir")


def test_tuple_entry_nested(tpl):
    tpl = TupleEntry.wrapping(tpl, str_sep=".")
    expected = ((3,), (4, 5), (6,))
    assert tpl.from_toml("3.4,5.6") == expected
    assert tpl.from_toml(["3", [4, 5], [6]]) == expected
    assert tpl.from_toml(expected) == expected
    assert tpl.to_toml(expected) == expected


def test_tuple_entry_cli(conf, climan):
    assert conf.sec.tpl == ()
    climan.parse_args(shsplit("--tpl 1,2,3"))
    assert conf.sec.tpl == (1, 2, 3)


def test_maybe_entry_from_none(maybe_path, maybe_int):
    assert maybe_path.from_toml(None) is None
    assert maybe_path.from_toml("") is None
    assert maybe_int.from_toml(None) is None
    assert maybe_int.from_toml("none") is None


def test_maybe_entry_from_val(maybe_path, maybe_int):
    assert maybe_path.from_toml("foo/bar") == Path("foo/bar")
    assert maybe_path.from_toml(Path()) == Path()
    assert maybe_int.from_toml("2") == 2
    assert maybe_int.from_toml(42) == 42


def test_maybe_entry_none_to_toml(maybe_path, maybe_int):
    assert maybe_path.to_toml(None) == ""
    assert maybe_int.to_toml(None) == "none"


def test_maybe_entry_val_to_toml(maybe_path, maybe_int):
    assert maybe_path.to_toml(Path("foo/bar")) == "foo/bar"
    assert maybe_int.to_toml(5) == 5


def test_maybe_entry_cli_empty(conf, climan):
    conf.sec.mfloat = 1.0
    conf.sec.mpath = Path()
    climan.parse_args(shsplit("--mfloat --mpath"))
    assert conf.sec.mfloat is None
    assert conf.sec.mpath is None


def test_maybe_entry_cli_val(conf, climan):
    climan.parse_args(shsplit("--mfloat 3.14 --mpath foo/bar"))
    assert conf.sec.mfloat == 3.14
    assert conf.sec.mpath == Path("foo") / "bar"



EXPECTED_ZSH = r"""#compdef cmd

function _cmd {
local line
_arguments -C \
"1:Commands:((sectionB\:'sectionB subcmd help'))" \
+ '(help)' \
'--help[show help message]' \
'-h[show help message]' \
+ '(optA)' \
'--optA=[AA]: :( )' \
'-a=[AA]: :( )' \
+ '(optB)' \
'--optB=[AB]: :( )' \
+ '(optC)' \
'--optC=[AC]: :( )' \
+ '(optBool)' \
'-optBool[Abool]' \
'+optBool[Abool]' \
'-o[Abool]' \
'+o[Abool]' \
'*::arg:->args'
case $line[1] in
sectionB) _cmd_sectionB ;;
esac
}

function _cmd_sectionB {
_arguments \
+ '(help)' \
'--help[show help message]' \
'-h[show help message]' \
+ '(optA)' \
'--optA=[BA]: :( )' \
+ '(optB)' \
'--optB=[BB]: :( )' \
+ '(optBool)' \
'-optBool[Bbool]' \
'+optBool[Bbool]' \
'-o[Bbool]' \
'+o[Bbool]' \
}
"""


EXPECTED_BASH = """_cmd() {
COMPREPLY=()
local cur=${COMP_WORDS[COMP_CWORD]}

local options="-h --help --optA -a --optB --optC -optBool +optBool -o +o"

local commands="sectionB"
declare -A suboptions
suboptions[sectionB]="-h --help --optA --optB -optBool +optBool -o +o"
if [[ "${COMP_LINE}" == *" sectionB "* ]] ; then
COMPREPLY=( `compgen -W "${suboptions[sectionB]}" -- ${cur}` )
elif [[ ${cur} == -* ]] ; then
COMPREPLY=( `compgen -W "${options}" -- ${cur}`)
else
COMPREPLY=( `compgen -W "${commands}" -- ${cur}`)
fi
}

complete -F _cmd cmd
"""


def test_bash_complete_file(climan, tmp_path):
    script_bash = tmp_path / "cmd.sh"
    climan.bash_complete(script_bash, "cmd")
    produced_bash = script_bash.read_text()
    assert produced_bash == EXPECTED_BASH


def test_zsh_complete_file(climan, tmp_path):
    script_zsh = tmp_path / "_cmd.sh"
    climan.zsh_complete(script_zsh, "cmd", force_grouping=True)
    produced_zsh = script_zsh.read_text()
    assert produced_zsh == EXPECTED_ZSH



"""Main classes to define your configuration."""

from __future__ import annotations

from dataclasses import dataclass, asdict, fields, field, Field
from os import PathLike
from pathlib import Path
from typing import (
    get_type_hints,
    TypeVar, Generic, Callable, Optional, Dict, Any, Type, Union, Mapping,
    ContextManager
)

import toml

from . import _internal

T = TypeVar("T")


@dataclass(frozen=True)
class Entry(Generic[T]):
    """Metadata of configuration options.

    Attributes:
        val: default value. Use :attr:`val_toml` or :attr:`val_factory` instead
            if it is mutable.
        val_toml: default value as a TOML value. :attr:`from_toml` is required.
            The call to the latter is wrapped in a function to avoid issues if
            the obtained value is mutable.
        val_factory: default value wrapped in a function, this is useful if the
            default value is mutable. This can be used to set a default value
            of `None`: `val_factory=lambda: None`.
        doc: short description of the option.
        from_toml: function to cast a value that can be represented as a TOML
            value to the type of the entry. When set, this is always called
            when reading a TOML file and command line arguments. Make sure you
            implement all the cases you want to support and raise a `TypeError`
            in other cases.
        to_toml: function to cast the entry to a type that can be represented
            as a TOML value. This is called when writing :class:`ConfigBase`
            instances to a TOML file via :meth:`ConfigBase.to_file_`.
        in_file: whether the option can be set in the config file.
        in_cli: whether the option is a command line argument.
        cli_short: short version of the command line argument.
        cli_kwargs: keyword arguments fed to
            :meth:`argparse.ArgumentParser.add_argument` during the
            construction of the command line arguments parser.
        cli_zsh_comprule: completion rule for ZSH shell.
    """

    val: Optional[T] = None
    val_toml: Optional[str] = None
    val_factory: Optional[Callable[[], T]] = None
    doc: str = ""
    from_toml: Optional[Callable[[object], T]] = None
    to_toml: Optional[Callable[[T], object]] = None
    in_file: bool = True
    in_cli: bool = True
    cli_short: Optional[str] = None
    cli_kwargs: Dict[str, Any] = field(default_factory=dict)
    cli_zsh_comprule: Optional[str] = ''

    def field(self) -> T:
        """Produce a :class:`dataclasses.Field` from the entry."""
        non_none_cout = (int(self.val is not None) +
                         int(self.val_toml is not None) +
                         int(self.val_factory is not None))
        if non_none_cout != 1:
            raise ValueError(
                "Exactly one of val, val_toml, and val_factory should be set.")

        if self.val is not None:
            return field(default=self.val, metadata=dict(loam_entry=self))
        if self.val_factory is not None:
            func = self.val_factory
        else:
            if self.from_toml is None:
                raise ValueError("Need `from_toml` to use val_toml")

            def func() -> T:
                # TYPE SAFETY: previous checks ensure this is valid
                return self.from_toml(self.val_toml)  # type: ignore

        return field(default_factory=func, metadata=dict(loam_entry=self))


def entry(
    val: Optional[T] = None,
    val_toml: Optional[str] = None,
    val_factory: Optional[Callable[[], T]] = None,
    doc: str = "",
    from_toml: Optional[Callable[[object], T]] = None,
    to_toml: Optional[Callable[[T], object]] = None,
    in_file: bool = True,
    in_cli: bool = True,
    cli_short: Optional[str] = None,
    cli_kwargs: Dict[str, Any] = None,
    cli_zsh_comprule: Optional[str] = '',
) -> T:
    """Build Entry(...).field()."""
    if cli_kwargs is None:
        cli_kwargs = {}
    return Entry(
        val=val,
        val_toml=val_toml,
        val_factory=val_factory,
        doc=doc,
        from_toml=from_toml,
        to_toml=to_toml,
        in_file=in_file,
        in_cli=in_cli,
        cli_short=cli_short,
        cli_kwargs=cli_kwargs,
        cli_zsh_comprule=cli_zsh_comprule,
    ).field()


@dataclass(frozen=True)
class Meta(Generic[T]):
    """Group several metadata of configuration entry.

    Attributes:
        fld: :class:`dataclasses.Field` object from the underlying metadata.
        entry: the metadata from the loam API.
        type_hint: type hint resolved as a class. If the type hint could not
            be resolved as a class, this is merely :class:`object`.
    """

    fld: Field[T]
    entry: Entry[T]
    type_hint: Type[T]


@dataclass
class Section:
    """Base class for a configuration section.

    This implements :meth:`__post_init__`. If your subclass also implement
    it, please call the parent implementation.
    """

    @classmethod
    def _type_hints(cls) -> Dict[str, Any]:
        return get_type_hints(cls)

    def __post_init__(self) -> None:
        self._loam_meta: Dict[str, Meta] = {}
        thints = self._type_hints()
        for fld in fields(self):
            meta = fld.metadata.get("loam_entry", Entry())
            thint = thints[fld.name]
            if not isinstance(thint, type):
                thint = object
            self._loam_meta[fld.name] = Meta(fld, meta, thint)
            current_val = getattr(self, fld.name)
            if not isinstance(current_val, thint):
                self.cast_and_set_(fld.name, current_val)

    def meta_(self, entry_name: str) -> Meta:
        """Metadata for the given entry name."""
        return self._loam_meta[entry_name]

    def cast_and_set_(self, field_name: str, value_to_cast: object) -> None:
        """Set an option from the string representation of the value.

        This uses :meth:`Entry.from_toml` (if present) to cast the given value.
        If :meth:`Entry.from_toml` is not present and the type of
        `value_to_cast` do not match the type hint, this calls the type hint to
        attempt a cast. This should only be used when setting an option to a
        value whose type cannot be controlled. Wherever possible, directly set
        the option value with the correct type instead of calling this method.
        """
        meta = self._loam_meta[field_name]
        if meta.entry.from_toml is not None:
            value = meta.entry.from_toml(value_to_cast)
        elif not isinstance(value_to_cast, meta.type_hint):
            try:
                value = meta.type_hint(value_to_cast)
            except Exception:
                raise TypeError(
                    f"Couldn't cast {value_to_cast!r} to a {meta.type_hint}, "
                    f"you might need to specify `from_toml` for {field_name}.")
        else:
            value = value_to_cast
        setattr(self, field_name, value)

    def context_(self, **options: Any) -> ContextManager[None]:
        """Enter a context with locally changed option values.

        This context is reusable but not reentrant.
        """
        return _internal.SectionContext(self, options)

    def update_from_dict_(self, options: Mapping[str, object]) -> None:
        """Update options from a mapping, casting values as needed."""
        for opt, val in options.items():
            self.cast_and_set_(opt, val)


TConfig = TypeVar("TConfig", bound="ConfigBase")


@dataclass
class ConfigBase:
    """Base class for a full configuration."""

    @classmethod
    def _type_hints(cls) -> Dict[str, Any]:
        return get_type_hints(cls)

    @classmethod
    def default_(cls: Type[TConfig]) -> TConfig:
        """Create a configuration with default values."""
        thints = cls._type_hints()
        sections = {}
        for fld in fields(cls):
            thint = thints[fld.name]
            if not (isinstance(thint, type) and issubclass(thint, Section)):
                raise TypeError(
                    f"Could not resolve type hint of {fld.name} to a Section "
                    f"(got {thint})")
            sections[fld.name] = thint()
        return cls(**sections)

    def update_from_file_(self, path: Union[str, PathLike]) -> None:
        """Update configuration from toml file."""
        pars = toml.load(Path(path))
        # only keep entries for which in_file is True
        pars = {
            sec_name: {
                opt: val
                for opt, val in section.items()
                if getattr(self, sec_name).meta_(opt).entry.in_file
            }
            for sec_name, section in pars.items()
        }
        self.update_from_dict_(pars)

    def update_from_dict_(
        self, options: Mapping[str, Mapping[str, Any]]
    ) -> None:
        """Update configuration from a dictionary."""
        for sec, opts in options.items():
            section: Section = getattr(self, sec)
            section.update_from_dict_(opts)

    def to_file_(
        self, path: Union[str, PathLike], exist_ok: bool = True
    ) -> None:
        """Write configuration in toml file."""
        path = Path(path)
        if not exist_ok and path.is_file():
            raise RuntimeError(f"{path} already exists")
        path.parent.mkdir(parents=True, exist_ok=True)
        dct = asdict(self)
        to_dump: Dict[str, Dict[str, Any]] = {}
        for sec_name, sec_dict in dct.items():
            to_dump[sec_name] = {}
            section: Section = getattr(self, sec_name)
            for fld in fields(section):
                entry = section.meta_(fld.name).entry
                if not entry.in_file:
                    continue
                value = sec_dict[fld.name]
                if entry.to_toml is not None:
                    value = entry.to_toml(value)
                to_dump[sec_name][fld.name] = value
            if not to_dump[sec_name]:
                del to_dump[sec_name]
        with path.open('w') as pf:
            toml.dump(to_dump, pf)



from shlex import split
import pytest
import loam.error
import loam.cli


def test_parse_no_args(conf, climan):
    climan.parse_args([])
    assert conf == conf.default_()


def test_parse_nosub_common_args(conf, climan):
    climan.parse_args(split('--optA 42'))
    assert conf.sectionA.optA == 42
    assert conf.sectionB.optA == 4


def test_parse_short_nosub_common_args(conf, climan):
    climan.parse_args(split('-a 42'))
    assert conf.sectionA.optA == 42
    assert conf.sectionB.optA == 4


def test_parse_switch_nosub_common_args(conf, climan):
    climan.parse_args(split('-optBool'))
    assert conf.sectionA.optBool is False
    assert conf.sectionB.optBool is False


def test_parse_switch_short_nosub_common_args(conf, climan):
    climan.parse_args(split('-o'))
    assert conf.sectionA.optBool is False
    assert conf.sectionB.optBool is False


def test_parse_sub_common_args(conf, climan):
    climan.parse_args(split('sectionB --optA 42'))
    assert conf.sectionA.optA == 1
    assert conf.sectionB.optA == 42


def test_parse_switch_sub_common_args(conf, climan):
    climan.parse_args(split('sectionB +optBool'))
    assert conf.sectionA.optBool is True
    assert conf.sectionB.optBool is True


def test_parse_switch_short_sub_common_args(conf, climan):
    climan.parse_args(split('sectionB +o'))
    assert conf.sectionA.optBool is True
    assert conf.sectionB.optBool is True


def test_parse_no_sub_only_args(conf, climan):
    climan.parse_args(split('--optC 42 sectionB'))
    assert conf.sectionA.optC == 3
    assert conf.sectionB.optC == 6


def test_parse_not_conf_cmd_args(climan):
    with pytest.raises(SystemExit):
        climan.parse_args(split('sectionB --optC 42'))


def test_build_climan_invalid_sub(conf):
    with pytest.raises(loam.error.SubcmdError):
        loam.cli.CLIManager(conf, **{'1invalid_sub': loam.cli.Subcmd('')})



"""Various helper functions and classes."""

from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
import subprocess
import shlex
import typing

from . import _internal
from .base import Entry, Section

if typing.TYPE_CHECKING:
    from typing import Optional, Union, Type
    from os import PathLike
    from .base import ConfigBase


def path_entry(
    path: Union[str, PathLike],
    doc: str,
    in_file: bool = True,
    in_cli: bool = True,
    cli_short: Optional[str] = None,
    cli_zsh_only_dirs: bool = False,
    cli_zsh_comprule: Optional[str] = None
) -> Path:
    """Define a path option.

    This creates a path option. See :class:`loam.base.Entry` for the meaning of
    the arguments. By default, the zsh completion rule completes any file. You
    can switch this to only directories with the `cli_zsh_only_dirs` option, or
    set your own completion rule with `cli_zsh_comprule`.
    """
    if cli_zsh_comprule is None:
        cli_zsh_comprule = "_files"
        if cli_zsh_only_dirs:
            cli_zsh_comprule += " -/"
    return Entry(
        val=Path(path),
        doc=doc,
        # TYPE SAFETY: Path behaves as needed
        from_toml=Path,  # type: ignore
        to_toml=str,
        in_file=in_file,
        in_cli=in_cli,
        cli_short=cli_short,
        cli_zsh_comprule=cli_zsh_comprule,
    ).field()


def switch_opt(default: bool, shortname: Optional[str],
               doc: str) -> bool:
    """Define a switchable option.

    This creates a boolean option. If you use it in your CLI, it can be
    switched on and off by prepending + or - to its name: +opt / -opt.

    Args:
        default: the default value of the swith option.
        shortname: short name of the option, no shortname will be used if set
            to None.
        doc: short description of the option.
    """
    return Entry(
        val=default, doc=doc, cli_short=shortname,
        cli_kwargs=dict(action=_internal.Switch), cli_zsh_comprule=None
    ).field()


def command_flag(doc: str, shortname: Optional[str] = None) -> bool:
    """Define a command line flag.

    The corresponding option is set to true if it is passed as a command line
    option.  This is similar to :func:`switch_opt`, except the option is not
    available from config files.  There is therefore no need for a mechanism to
    switch it off from the command line.

    Args:
        doc: short description of the option.
        shortname: short name of the option, no shortname will be used if set
            to None.
    """
    return Entry(  # previously, default value was None. Diff in cli?
        val=False, doc=doc, in_file=False, cli_short=shortname,
        cli_kwargs=dict(action="store_true"), cli_zsh_comprule=None
    ).field()


@dataclass
class ConfigSection(Section):
    """A configuration section handling config files."""

    create: bool = command_flag("create global config file")
    update: bool = command_flag("add missing entries to config file")
    edit: bool = command_flag("open config file in a text editor")
    editor: str = Entry(val="vim", doc='text editor').field()


def config_cmd_handler(
        config: Union[ConfigBase, Type[ConfigBase]],
        config_section: ConfigSection,
        config_file: Path,
) -> None:
    """Implement the behavior of a subcmd using config_conf_section.

    Args:
        config: the :class:`~loam.base.ConfigBase` to manage.
        config_section: a :class:`ConfigSection` set as desired.
        config_file: path to the config file.
    """
    if config_section.update:
        conf = config.default_()
        if config_file.exists():
            conf.update_from_file_(config_file)
        conf.to_file_(config_file)
    elif config_section.create or config_section.edit:
        config.default_().to_file_(config_file)
    if config_section.edit:
        subprocess.run(shlex.split('{} {}'.format(config_section.editor,
                                                  config_file)))



"""Definition of CLI manager."""
from __future__ import annotations
from dataclasses import fields
import argparse
import copy
import pathlib
import typing
import warnings
from types import MappingProxyType

from . import error, _internal

if typing.TYPE_CHECKING:
    from typing import Dict, List, Any, Optional, Mapping, TextIO, Union
    from argparse import ArgumentParser, Namespace
    from os import PathLike
    from .base import Section, ConfigBase


BLK = ' \\\n'  # cutting line in scripts


def _names(section: Section, option: str) -> List[str]:
    """List of cli strings for a given option."""
    entry = section.meta_(option).entry
    action = entry.cli_kwargs.get('action')
    if action is _internal.Switch:
        names = [f'-{option}', f'+{option}']
        short = entry.cli_short
        if short is not None:
            names.append(f'-{short}')
            names.append(f'+{short}')
    else:
        names = [f'--{option}']
        short = entry.cli_short
        if short is not None:
            names.append(f'-{short}')
    return names


class Subcmd:
    """Metadata of sub commands.

    Attributes:
        help: short description of the sub command.
        sections: configuration sections used by the subcommand.
        defaults: default value of options associated to the subcommand.
    """

    def __init__(self, help_msg: str, *sections: str, **defaults: Any):
        self.help = help_msg
        self.sections = sections
        self.defaults = defaults


class CLIManager:
    """CLI manager.

    Args:
        config_: the :class:`~loam.base.ConfigBase` holding option definitions.
        common_: special subcommand, used to define the general description
            of the CLI tool as well as configuration sections used by every
            subcommand.
        bare_: special subcommand, use it to define the configuration
            sections that should be used when you call your CLI tool
            without any subcommand.
        subcmds: all the subcommands of your CLI tool. The name of each
            *subcommand* is the name of the keyword argument passed on to
            this function.
    """

    def __init__(self, config_: ConfigBase,
                 common_: Optional[Subcmd] = None,
                 bare_: Optional[Subcmd] = None,
                 **subcmds: Subcmd):
        self._conf = config_
        self._subcmds = {}
        for sub_name, sub_meta in subcmds.items():
            if sub_name.isidentifier():
                self._subcmds[sub_name] = sub_meta
            else:
                raise error.SubcmdError(sub_name)
        self._common = common_ if common_ is not None else Subcmd('')
        self._bare = bare_
        # dict of dict [command][option] = section
        self._opt_cmds: Dict[str, Dict[str, str]] = {}
        # same as above but for bare command only [option] = section
        self._opt_bare: Dict[str, str] = {}
        if self.bare is not None:
            self._cmd_opts_solver(None)
        for cmd_name in self.subcmds:
            self._opt_cmds[cmd_name] = {}
            self._cmd_opts_solver(cmd_name)
        self._parser = self._build_parser()

    @property
    def common(self) -> Subcmd:
        """Subcmd describing sections common to all subcommands."""
        return self._common

    @property
    def bare(self) -> Optional[Subcmd]:
        """Subcmd used when the CLI tool is invoked without subcommand."""
        return self._bare

    @property
    def subcmds(self) -> Mapping[str, Subcmd]:
        """Subcommands description."""
        return MappingProxyType(self._subcmds)

    def sections_list(self, cmd: Optional[str] = None) -> List[str]:
        """List of config sections used by a command.

        Args:
            cmd: command name, set to ``None`` or ``''`` for the bare command.

        Returns:
            list of configuration sections used by that command.
        """
        sections = list(self.common.sections)
        if not cmd:
            if self.bare is not None:
                sections.extend(self.bare.sections)
                return sections
            return []
        sections.extend(self.subcmds[cmd].sections)
        if hasattr(self._conf, cmd):
            sections.append(cmd)
        return sections

    def _cmd_opts_solver(self, cmd_name: Optional[str]) -> None:
        """Scan options related to one command and enrich _opt_cmds."""
        sections = self.sections_list(cmd_name)
        cmd_dict = self._opt_cmds[cmd_name] if cmd_name else self._opt_bare
        for sct in reversed(sections):
            section: Section = getattr(self._conf, sct)
            for fld in fields(section):
                opt = fld.name
                if not section.meta_(opt).entry.in_cli:
                    continue
                if opt not in cmd_dict:
                    cmd_dict[opt] = sct
                else:
                    warnings.warn(
                        'Command <{0}>: {1}.{2} shadowed by {3}.{2}'.format(
                            cmd_name, sct, opt, cmd_dict[opt]),
                        error.LoamWarning, stacklevel=4)

    def _add_options_to_parser(self, opts_dict: Mapping[str, str],
                               parser: ArgumentParser) -> None:
        """Add options to a parser."""
        for opt, sct in opts_dict.items():
            section: Section = getattr(self._conf, sct)
            entry = section.meta_(opt).entry
            kwargs = copy.deepcopy(entry.cli_kwargs)
            action = kwargs.get('action')
            if action is _internal.Switch:
                kwargs.update(nargs=0)
            kwargs.update(help=entry.doc)
            kwargs.setdefault('default', getattr(section, opt))
            parser.add_argument(*_names(section, opt), **kwargs)

    def _build_parser(self) -> ArgumentParser:
        """Build command line argument parser.

        Returns:
            the command line argument parser.
        """
        main_parser = argparse.ArgumentParser(description=self.common.help,
                                              prefix_chars='-+')

        self._add_options_to_parser(self._opt_bare, main_parser)
        main_parser.set_defaults(**self.common.defaults)
        if self.bare is not None:
            main_parser.set_defaults(**self.bare.defaults)

        subparsers = main_parser.add_subparsers(dest='loam_sub_name')
        for cmd_name, meta in self.subcmds.items():
            kwargs = {'prefix_chars': '+-', 'help': meta.help}
            dummy_parser = subparsers.add_parser(cmd_name, **kwargs)
            self._add_options_to_parser(self._opt_cmds[cmd_name], dummy_parser)
            dummy_parser.set_defaults(**meta.defaults)

        return main_parser

    def parse_args(self, arglist: Optional[List[str]] = None) -> Namespace:
        """Parse arguments and update options accordingly.

        Args:
            arglist: list of arguments to parse. If set to None,
                ``sys.argv[1:]`` is used.

        Returns:
            the argument namespace returned by the
            :class:`argparse.ArgumentParser`.
        """
        args = self._parser.parse_args(args=arglist)
        sub_cmd = args.loam_sub_name
        if sub_cmd is None:
            for opt, sct in self._opt_bare.items():
                section: Section = getattr(self._conf, sct)
                val = getattr(args, opt, None)
                section.cast_and_set_(opt, val)
        else:
            for opt, sct in self._opt_cmds[sub_cmd].items():
                section = getattr(self._conf, sct)
                val = getattr(args, opt, None)
                section.cast_and_set_(opt, val)
        return args

    def _zsh_comp_command(self, zcf: TextIO, cmd: Optional[str],
                          grouping: bool, add_help: bool = True) -> None:
        """Write zsh _arguments compdef for a given command.

        Args:
            zcf: zsh compdef file.
            cmd: command name, set to None or '' for bare command.
            grouping: group options (zsh>=5.4).
            add_help: add an help option.
        """
        if add_help:
            if grouping:
                print("+ '(help)'", end=BLK, file=zcf)
            print("'--help[show help message]'", end=BLK, file=zcf)
            print("'-h[show help message]'", end=BLK, file=zcf)
        # could deal with duplicate by iterating in reverse and keep set of
        # already defined opts.
        no_comp = ('store_true', 'store_false')
        cmd_dict = self._opt_cmds[cmd] if cmd else self._opt_bare
        for opt, sct in cmd_dict.items():
            section: Section = getattr(self._conf, sct)
            entry = section.meta_(opt).entry
            comprule = entry.cli_zsh_comprule
            if entry.cli_kwargs.get('action') == 'append':
                grpfmt, optfmt = "+ '{}'", "'*{}[{}]{}'"
                if comprule is None:
                    comprule = ''
            else:
                grpfmt, optfmt = "+ '({})'", "'{}[{}]{}'"
            if entry.cli_kwargs.get('action') in no_comp \
               or entry.cli_kwargs.get('nargs') == 0:
                comprule = None
            if comprule is None:
                compstr = ''
            elif comprule == '':
                optfmt = optfmt.replace('[', '=[')
                compstr = ': :( )'
            else:
                optfmt = optfmt.replace('[', '=[')
                compstr = f': :{comprule}'
            if grouping:
                print(grpfmt.format(opt), end=BLK, file=zcf)
            for name in _names(section, opt):
                print(optfmt.format(name, entry.doc.replace("'", "'\"'\"'"),
                                    compstr), end=BLK, file=zcf)

    def zsh_complete(self, path: Union[str, PathLike], cmd: str, *cmds: str,
                     sourceable: bool = False,
                     force_grouping: bool = False) -> None:
        """Write zsh compdef script.

        Args:
            path: desired path of the compdef script.
            cmd: command name that should be completed.
            cmds: extra command names that should be completed.
            sourceable: if True, the generated file will contain an explicit
                call to ``compdef``, which means it can be sourced to activate
                CLI completion.
            force_grouping: if True, assume zsh supports grouping of options.
                Otherwise, loam will attempt to check whether zsh >= 5.4.
        """
        grouping = force_grouping or _internal.zsh_version() >= (5, 4)
        path = pathlib.Path(path)
        firstline = ['#compdef', cmd]
        firstline.extend(cmds)
        subcmds = list(self.subcmds.keys())
        with path.open('w') as zcf:
            print(*firstline, end='\n\n', file=zcf)
            # main function
            print(f'function _{cmd} {{', file=zcf)
            print('local line', file=zcf)
            print('_arguments -C', end=BLK, file=zcf)
            if subcmds:
                # list of subcommands and their description
                substrs = [rf"{sub}\:'{self.subcmds[sub].help}'"
                           for sub in subcmds]
                print('"1:Commands:(({}))"'.format(' '.join(substrs)),
                      end=BLK, file=zcf)
            self._zsh_comp_command(zcf, None, grouping)
            if subcmds:
                print("'*::arg:->args'", file=zcf)
                print('case $line[1] in', file=zcf)
                for sub in subcmds:
                    print(f'{sub}) _{cmd}_{sub} ;;', file=zcf)
                print('esac', file=zcf)
            print('}', file=zcf)
            # all subcommand completion handlers
            for sub in subcmds:
                print(f'\nfunction _{cmd}_{sub} {{', file=zcf)
                print('_arguments', end=BLK, file=zcf)
                self._zsh_comp_command(zcf, sub, grouping)
                print('}', file=zcf)
            if sourceable:
                print(f'\ncompdef _{cmd} {cmd}', *cmds, file=zcf)

    def _bash_comp_command(self, cmd: Optional[str],
                           add_help: bool = True) -> List[str]:
        """Build a list of all options for a given command.

        Args:
            cmd: command name, set to None or '' for bare command.
            add_help: add an help option.

        Returns:
            list of CLI options strings.
        """
        out = ['-h', '--help'] if add_help else []
        cmd_dict = self._opt_cmds[cmd] if cmd else self._opt_bare
        for opt, sct in cmd_dict.items():
            section: Section = getattr(self._conf, sct)
            out.extend(_names(section, opt))
        return out

    def bash_complete(self, path: Union[str, PathLike], cmd: str,
                      *cmds: str) -> None:
        """Write bash complete script.

        Args:
            path: desired path of the complete script.
            cmd: command name that should be completed.
            cmds: extra command names that should be completed.
        """
        path = pathlib.Path(path)
        subcmds = list(self.subcmds.keys())
        with path.open('w') as bcf:
            # main function
            print(f'_{cmd}() {{', file=bcf)
            print('COMPREPLY=()', file=bcf)
            print(r'local cur=${COMP_WORDS[COMP_CWORD]}', end='\n\n', file=bcf)
            optstr = ' '.join(self._bash_comp_command(None))
            print(f'local options="{optstr}"', end='\n\n', file=bcf)
            if subcmds:
                print('local commands="{}"'.format(' '.join(subcmds)),
                      file=bcf)
                print('declare -A suboptions', file=bcf)
            for sub in subcmds:
                optstr = ' '.join(self._bash_comp_command(sub))
                print(f'suboptions[{sub}]="{optstr}"', file=bcf)
            condstr = 'if'
            for sub in subcmds:
                print(condstr, r'[[ "${COMP_LINE}" == *"', sub, '"* ]] ; then',
                      file=bcf)
                print(r'COMPREPLY=( `compgen -W "${suboptions[', sub,
                      r']}" -- ${cur}` )', sep='', file=bcf)
                condstr = 'elif'
            print(condstr, r'[[ ${cur} == -* ]] ; then', file=bcf)
            print(r'COMPREPLY=( `compgen -W "${options}" -- ${cur}`)',
                  file=bcf)
            if subcmds:
                print(r'else', file=bcf)
                print(r'COMPREPLY=( `compgen -W "${commands}" -- ${cur}`)',
                      file=bcf)
            print('fi', file=bcf)
            print('}', end='\n\n', file=bcf)
            print(f'complete -F _{cmd} {cmd}', *cmds, file=bcf)
